﻿using UnityEngine;
using System.Collections;
using UnityEngine.SceneManagement;

[System.Serializable]
public class Hazard
{
    public GameObject hazard;
    public Vector3 spawnValues;
    public int hazardCount; //amount of hazards per wave
    public float spawnWait; //time inbetween hazard spawns
    public float waveWait; //time inbetween waves
}

[System.Serializable]
public class SharpHazard : Hazard
{

}

[System.Serializable]
public class GroundHazard : Hazard
{

}

[System.Serializable]
public class AirHazard : Hazard
{
    public AudioSource hazardSound;
}

public class GameController : MonoBehaviour
{
    //Hazard Types
    public SharpHazard sH;
    public GroundHazard gH;
    public AirHazard aH;

    //Enable Game Functions;
    public bool enableSharpHazards = true;
    public bool enableGroundHazards = true;
    public bool enableAirHazards = false;
    public bool enableTimeScore;

    //GUI Text elements
    public GUIText scoreText;
    public GUIText restartText;
    public GUIText gameOverText;
    public GUIText gameOverText_shadow;
    public GUIText livesText;
    public GUIText hitText;

    //Audio clips
    public AudioSource mainMusic;
    public AudioSource problemMusic;

    //Coroutines
    private Coroutine engineProblemRoutine;

    //MISC
    public float startWait = 3; //Time to get ready before the game starts
    public int timeScore = 1;   //Scored gained by staying alive
    public int lives = 3;       //Amount of times a player can get hit by a sharp hazard
    public bool coop = false;   //Idicate wether the game is played solo or coop

    private string[] curses = new string[] { "Ouch!", "Look out!",
                                     "Cor blimey!", "I will get ya' filthy scum!",
                                     "Come at me like a gentleman!" };
    private Vector3[] spawnPosArray; //Array used to generate specific spawn positions
    private int spawnPosIndex;       //Index used by the spawnPosArray
    private bool gameOver;           //True when the game is over
    private int score;
    private int difficultyLevel;
    
     

    void Start()
    {
        gameOver = false;
        restartText.text = "";
        gameOverText.text = "";
        gameOverText_shadow.text = "";
        livesText.text = "";
        hitText.text = "";
        adjustPlayerCount();
        score = 0;
        UpdateScore();

        //Start Audio
        mainMusic.volume = 1.0f;
        mainMusic.Play();
        problemMusic.volume = 0.0f;
        problemMusic.Play();

        //Start coroutines
        StartCoroutine(SpawnSharpWaves());
        StartCoroutine(SpawnGroundWaves());
        StartCoroutine(SpawnAirWaves());
        StartCoroutine(TimeScoreIncreaser());
    }

    void Update()
    {
        UpdateScore();
        UpdateLives();
        if (gameOver)
        {
            StopAllCoroutines();
            restartText.text = "Press 'R' for Restart or ESC to go to title screen";
            if (Input.GetKeyDown(KeyCode.R))
            {
                SceneManager.LoadScene("Main");
            }
            if (Input.GetKeyDown(KeyCode.Escape))
            {
                SceneManager.LoadScene("Title");
            }
        }
    }

    // ROUTINE FUNCTIONS \\

    /*
     * Resume the regular game routines after a probem.
     */
    public void pauseRegularRoutines()
    {
        enableSharpHazards = false;
        enableGroundHazards = false;
        enableAirHazards = false;
        enableTimeScore = false;
    }

    /*
     * Enables all regular game routines
     */
    public void startRegularRoutines()
    {
        enableSharpHazards = true;
        enableGroundHazards = true;
        enableAirHazards = true;
        enableTimeScore = true;
    }

    // GAME ROUTINES \\

    /*
     * Coroutine used to spawn waves of homing sharp hazards.
     * Each wave consists of an amount of hazards.
     * These hazards are spawned randomly at the right, left or top of the game.
     */
    IEnumerator SpawnSharpWaves()
    {
        yield return new WaitForSeconds(startWait);
        while (true)
        {
            if (enableSharpHazards)
            {
                for (int i = 0; i < sH.hazardCount; i++)
                {
                    //Calculate random spawn location on right, left and top
                    Vector3 spawnPosLeft = new Vector3(-sH.spawnValues.x, sH.spawnValues.y, Random.Range(-sH.spawnValues.z, sH.spawnValues.z + 2));
                    //Vector3 spawnPosTop = new Vector3(Random.Range(-sH.spawnValues.x, sH.spawnValues.x), sH.spawnValues.y, sH.spawnValues.z + 1);
                    Vector3 spawnPosRight = new Vector3(sH.spawnValues.x, sH.spawnValues.y, Random.Range(-sH.spawnValues.z, sH.spawnValues.z + 2));

                    spawnPosArray = new Vector3[] { spawnPosRight, spawnPosLeft }; // spawnPosTop, };
                    spawnPosIndex = Random.Range(0, spawnPosArray.Length); //Choose random right, left or top
                    Instantiate(sH.hazard, spawnPosArray[spawnPosIndex], sH.hazard.GetComponent<Rigidbody>().rotation = Quaternion.Euler(0.0f, 0.0f, 0.0f)); //instantiate hazard with rotation based on spawn position.
                    Debug.Log(sH.GetType() + " spawned at position " + spawnPosArray[spawnPosIndex]);
                    yield return new WaitForSeconds(sH.spawnWait);
                }
                yield return new WaitForSeconds(sH.waveWait);
            }
            else
            {
                yield return null;
            }

        }
    }

    /*
     * Coroutine used to spawn waves of static ground hazards.
     * Each wave consists of an amount of hazards.
     * These hazards are spawned randomly at the bottom of the game.
     */
    IEnumerator SpawnGroundWaves()
    {
        yield return new WaitForSeconds(startWait);
        while (true)
        {
            if (enableGroundHazards)
            {
                for (int i = 0; i < gH.hazardCount; i++)
                {
                    //Calculate random spawn location
                    Vector3 spawnPosition = new Vector3(gH.spawnValues.x, gH.spawnValues.y, Random.Range(gH.spawnValues.z, gH.spawnValues.z - 3));
                    Instantiate(gH.hazard, spawnPosition, gH.hazard.GetComponent<Rigidbody>().rotation = Quaternion.Euler(90.0f, 0.0f, 0.0f)); //instantiate with fixed angle
                    Debug.Log(gH.GetType() + " spawned at position " + spawnPosition);
                    yield return new WaitForSeconds(gH.spawnWait);
                }
                yield return new WaitForSeconds(gH.waveWait);
            } else
            {
                yield return null;
            }
        }
    }

    /*
     * Coroutine used to spawn waves of static air hazards.
     * Each wave consists of an amount of hazards.
     * These hazards are spawned randomly at the top of the game.
     */
    IEnumerator SpawnAirWaves()
    {
        yield return new WaitForSeconds(startWait);
        while (true)
        {
            if (enableAirHazards)
            {
                for (int i = 0; i < gH.hazardCount; i++)
                {
                    aH.hazardSound.Play();
                    yield return new WaitForSeconds(7);
                    //Calculate random spawn location
                    Vector3 spawnPosition = new Vector3(aH.spawnValues.x, aH.spawnValues.y, Random.Range(aH.spawnValues.z, aH.spawnValues.z + 1.5f));

                    Object x = Instantiate(aH.hazard, spawnPosition, aH.hazard.GetComponent<Rigidbody>().rotation = Quaternion.Euler(180.0f, -90.0f, -90.0f)); //instantiate with fixed angle
                    Debug.Log(aH.GetType() + " spawned at position " + spawnPosition);
                    yield return new WaitForSeconds(gH.spawnWait);
                }
                yield return new WaitForSeconds(gH.waveWait);
            }
            else
            {
                yield return null;
            }
        }
    }

    /*
     * Coroutine used to add score based on the amount of time the player is alive.
     * When the player is game over, the scoring stops.
     */
    IEnumerator TimeScoreIncreaser()
    {
        while (true)
        {
            if (enableTimeScore)
            {
                score += timeScore;
                yield return new WaitForSeconds(1);
            }
            else
            {
                yield return null;
            }
            
        }
    }

    /*
     * Increases the difficulty of the game by altering the hazard values
     */
    public void  increaseHazardDifficulty() {
        //Sharp Hazards
        sH.hazardCount++;
        sH.waveWait += 0.5f;

        //Ground Hazards
        gH.hazardCount++;
        if (gH.waveWait > 5)
        {
            gH.waveWait--;
        }
        
        //Flying Hazards
        if (difficultyLevel == 1)
        {
            enableAirHazards = true;
        }
        if (difficultyLevel > 1)
        {
            if (sH.waveWait > 3)
            {
                sH.waveWait--;
            }        
        }

    }

    // SCORE FUNCTIONS \\

    /*
     * Function to increase the score value by a certain value.
     */
    public void AddScore(int newScoreValue)
    {
        score += newScoreValue;
    }

    /*
     * Function to update the score displayed on the screen.
     */
    void UpdateScore()
    {
        scoreText.text = "Score: " + score;
    }

    // HIT FUNCTIONS \\

    /*
     * Function to update the lives displayed on the screen.
     */
    void UpdateLives()
    {
        livesText.text = "Lives: " + lives;
    }

    /*
     * Routine to display a curse for 3 seconds after being hit by a hazard
     */
    public IEnumerator hitLine()
    {
        hitText.text = curses[Random.Range(0, curses.Length)];
        yield return new WaitForSeconds(3);
        hitText.text = "";
    }

    // GAME OVER FUNCTIONS \\

    /*
     * Function to instantiate the game over state.
     */
    public void GameOver()
    {
        {
            updateHighScore();
            gameOverText.text = "DEFLATED";
            gameOverText_shadow.text = "DEFLATED";
            gameOver = true;
        }
    }

    /*
     * Auxiliary function to get the gameOver variable.
     */
    public bool getGameOver()
    {
        return gameOver;
    }

    /*
     * Update highscore by reading the current and writing the higher if it is the case
     */
    public void updateHighScore()
    {
        int currHigh = System.Int32.Parse(System.IO.File.ReadAllText("Assets/Other/HighScore.txt"));

        if (score > currHigh)
        {
            System.IO.File.WriteAllText("Assets/Other/HighScore.txt", "" + score);
        }
    }

    // MUSIC FUNCTIONS\\

    /*
    * switch between main music and problem music
    */
    public IEnumerator switchMusic()
    {
        float t = 1.0f;
        if (mainMusic.volume == 1.0f)
        {

            while (t >= 0.0f)
            {
                t -= Time.deltaTime;
                mainMusic.volume = t;
                problemMusic.volume = 1.0f - t;
                yield return null;
            }

        }
        else
        {
            while (t >= 0.0f)
            {
                t -= Time.deltaTime;
                problemMusic.volume = t;
                mainMusic.volume = 1.0f - t;
                yield return null;
            }
        }
    }

    /*
     * Fade out current music and stop playing it.
     */
    public IEnumerator fadeOutAndStop(AudioSource music)
    {
        float t = 1.0f;
        if (music.volume != 0.0f)
        {
            while (t >= 0.0f)
            {
                t -= Time.deltaTime;
                music.volume = t;
                yield return null;
            }
        }
        music.Stop();
    }

    // CO-OP \\
    
    /*
     * Reads the amount of players written in the text file by the title screen
     * and adjust the coop indicator accordingly.
     */
    private void adjustPlayerCount()
    {
        int amount = System.Int32.Parse(System.IO.File.ReadAllText("Assets/Other/PlayerCount.txt"));

        if (amount == 1) //Solo
        {
            coop = false;
        }
        else if (amount == 2) //Co-op
        {
            coop = true;
        }
        else
        {
            throw new System.Exception("Forbidden player count");
        }
    }
}
