﻿using UnityEngine;
using System.Collections;

[System.Serializable]
/*
 * class to maintain the player's boundary variables
 */
public class Boundary
{
    public float xMin, xMax, zMin, zMax;
}

public class PilotController : MonoBehaviour
{
    //MISC
    public Boundary boundary;
    public bool simpleControls; //use simple controls
    private GameController gameController;
    public GameObject balloonExplosion;

    //Vertical
    private float currVerSpeed = 0;
    public float maxVerSpeed = 5;
    public float verAcc = 1;
    public float verDec = 5;

    //Horizontal
    private float currHorSpeed = 0;
    public float maxHorSpeed = 3;
    public float horAcc = 1;
    public float horDec = 5;

    //Problem State
    public bool problemState = false;
    public float dropDownTime = 60;
    public float cycleEffectTime = 1.0f;
    public float reduceSpeedFactor = 0.5f;
    private float dropDownHeight;
    private bool dropCalculated;
    private KeyCode[] lastEffortCycle = new KeyCode[] {KeyCode.W, KeyCode.D, KeyCode.S, KeyCode.A};
    private bool cycleComplete;
    private float startTime;
    private float dropDownSpeed;

    void Start()
    {
        findGameController();

        //give player more problem solve time if playing solo
        if (!gameController.coop)
        {
            dropDownTime *= 1.5f;
        }
    }

    /*
     * allows the player to control to hot air balloon
     */
    void FixedUpdate()
    {
        //Regular controls when not in a problem state
        if (!problemState)
        {
            //when recovered from an engine problem, drop should be recalculated again for next problem
            if (dropCalculated)
            {
                dropCalculated = false;
            }
            //overwrite control by simple controls
            if (simpleControls)
            {
                float moveHorizontal = Input.GetAxis("Horizontal");
                float moveVertical = Input.GetAxis("Vertical");

                Vector3 movement = new Vector3(moveHorizontal, 0.0f, moveVertical);
                GetComponent<Rigidbody>().velocity = movement * maxVerSpeed;

                GetComponent<Rigidbody>().position = new Vector3(
                    Mathf.Clamp(GetComponent<Rigidbody>().position.x, boundary.xMin, boundary.xMax),
                    0.0f,
                    Mathf.Clamp(GetComponent<Rigidbody>().position.z, boundary.zMin, boundary.zMax)
                );
                return;
            }
            //calculates vertical accelaration of the balloon
            //when upwards is pressed and max speed is not yet reached: accelarate upwards
            if (Input.GetKey(KeyCode.W) && currVerSpeed < maxVerSpeed)
            {
                currVerSpeed = currVerSpeed + verAcc * Time.deltaTime;
            }
            else
            {
                //if key is released decelarate
                if (currVerSpeed > verDec * Time.deltaTime)
                {
                    currVerSpeed = currVerSpeed - verDec * Time.deltaTime;
                }
                //otherwise drop down (regular gravity does not apply)
                else
                {
                    currVerSpeed = -0.5f;
                }
            }

            //calculates horizontal accelaration of the balloon
            //when right is pressed and max speed is not yet reached: accelarate right
            if (Input.GetKey(KeyCode.A) && currHorSpeed < maxHorSpeed)
            {
                currHorSpeed = currHorSpeed - horAcc * Time.deltaTime;
            }
            //or when left is pressed and opposite max speed is not yet reached: accelarate left
            else if (Input.GetKey(KeyCode.D) && currHorSpeed > -maxHorSpeed)
            {
                currHorSpeed = currHorSpeed + horAcc * Time.deltaTime;
            }
            else
            {
                //if key is released decelarate from right speed
                if (currHorSpeed > horDec * Time.deltaTime)
                {
                    currHorSpeed = currHorSpeed - horDec * Time.deltaTime;
                }
                //if key is released decelarate from left speed
                else if (currHorSpeed < -horDec * Time.deltaTime)
                {
                    currHorSpeed = currHorSpeed + horDec * Time.deltaTime;
                }
                //otherwise no horizontal movement
                else
                {
                    currHorSpeed = 0;
                }
            }

        }
        //When in a problem state
        else
        {
            //Calculate falling speed based on current height
            if (!dropCalculated)
            {
                dropDownHeight = GetComponent<Rigidbody>().position.z - boundary.zMin;
                dropDownSpeed = -(dropDownHeight / dropDownTime);
                currVerSpeed = dropDownSpeed;
                currHorSpeed = 0;
                dropCalculated = true;
                Debug.Log("Drop from height: " + dropDownHeight + "\ndDrop speed: " + currVerSpeed);

                StartCoroutine(LastEffort());
            }
            //If a key sequence is completed in the last effort coroutine reduce falling speed for short time
            if (cycleComplete)
            {
                startTime = Time.time;
                cycleComplete = false;
            }
            if (Time.time - startTime < cycleEffectTime)
            {
                currVerSpeed = reduceSpeedFactor * dropDownSpeed; 
            }
            else
            {
                currVerSpeed = dropDownSpeed;
            }
        }

        //set the horizontal and vertical velocity (y equals 0, because 2D)
        GetComponent<Rigidbody>().velocity = new Vector3(currHorSpeed, 0, currVerSpeed);

        //Restrains the player from leaving the set boundaries
        GetComponent<Rigidbody>().position = new Vector3(
            Mathf.Clamp(GetComponent<Rigidbody>().position.x, boundary.xMin, boundary.xMax),
            0.0f,
            Mathf.Clamp(GetComponent<Rigidbody>().position.z, boundary.zMin, boundary.zMax));

        if (GetComponent<Rigidbody>().position.z == -6.0)
        {
            Destroy(gameObject);
            gameController.GameOver();
            Instantiate(balloonExplosion, transform.position, transform.rotation);
        }
    }

    /*
     * Coroutine to reduce falling speed by repeating a key sequence
     */
    public IEnumerator LastEffort()
    {
        int i = 0;
        int cycleLength = lastEffortCycle.Length;
        KeyCode currKey = lastEffortCycle[i];

        while (problemState)
        {
            yield return StartCoroutine(WaitForKey(currKey));
            if (++i == cycleLength)
            {
                cycleComplete = true;
                Debug.Log("W-D-S-A cycle completed: 0.5 * fallspeed");
            }
            i = i % cycleLength;
            currKey = lastEffortCycle[i];
        }
    }  

    /*
     * Coroutine which waits for a key input
     */
    private IEnumerator WaitForKey(KeyCode key)
{
    while (!Input.GetKeyUp(key))
    {
        yield return null;
    }
}

    /*
     * Auxiliary function to find the GameObject with the tag 'GameController' and
     * the attached GameController script.
     */
    public void findGameController()
    {
        GameObject gameControllerObject = GameObject.FindWithTag("GameController");
        if (gameControllerObject != null)
        {
            gameController = gameControllerObject.GetComponent<GameController>();
        }
        else if (gameControllerObject == null)
        {
            Debug.Log("Cannot find 'GameController' script");
        }
    }
}
