﻿using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using System.Linq;

public class RandomProblemGenerator : MonoBehaviour
{

    public GUIText problemText;
    public int problemSequenceLength = 3;
    public int rangeStart;
    public int rangeEnd;

    //All possible key values that will not interfere with other controls together with their display text
    private Dictionary<KeyCode, string> keysWithNames = new Dictionary<KeyCode, string>()
                                                        { {KeyCode.Q, "kool bore" }, {KeyCode.E, "fuel tank cap" }, {KeyCode.T, "side valve" },
                                                          {KeyCode.Y, "transportguard" }, {KeyCode.U, "cast iron camshaft" }, {KeyCode.I, "dual ball bearings" },
                                                          {KeyCode.O, "ignition system" }, {KeyCode.P, "high oil fill tube" }, {KeyCode.F, "pressure lubrication" },
                                                          {KeyCode.G, "foam air cleaner" }, {KeyCode.H, "dura-bore" }, {KeyCode.J, "lo-tone muffler" },
                                                          {KeyCode.K, "v-twin cylinder engine" }, {KeyCode.L, "gross torque" }, {KeyCode.Z, "float feed carburetor" },
                                                          {KeyCode.X, "fuel injection system" }, {KeyCode.C, "fuel filter" }, {KeyCode.V, "overhead valve" },
                                                          {KeyCode.B, "mechanical governor" }, {KeyCode.N, "compression release" }, {KeyCode.M, "super lo-tone muffler" },
                                                          {KeyCode.Alpha0, "debris management system" }, {KeyCode.Alpha1, "oil bath air cleaner" },
                                                          {KeyCode.Alpha2, "direct overhead valve" }, {KeyCode.Alpha3, "single cylinder engine" },
                                                          {KeyCode.Alpha4, "cartridge air cleaner" }, {KeyCode.Alpha5, "splash lubrication" },
                                                          {KeyCode.Alpha6, "anti-vibration system" }, {KeyCode.Alpha7, "starting system" },
                                                          {KeyCode.Alpha8, "forged crankshaft" }, {KeyCode.Alpha9, "weight measurer" },

                                                        };

    //All possible ordering values and their display text
    private Dictionary<string, string[]> ordersWithVerbs = new Dictionary<string, string[]>()
                                                         { {"inorder", new string[]{"booster", "catalytic converter", "radio channel", "exhaust system",
                                                                                    "oxygen sensor", "emergency exit", "altimeter", "propane burner",
                                                                                    "parachute valve cord", "propane fuel tanks" }},
                                                           {"ascending", new string[] {"cylinder heads", "piston liners", "air boxes",
                                                                                       "varlometers", "sump pans", "distributor caps",
                                                                                       "ignition leads", "spark plugs", "compass", "boost pressure sensors"}},
                                                           {"descending", new string[] {"piston engine valves", "intercooler", "turbocharger gaskets",
                                                                                        "pollen filters", "oil seals", "flywheel", "diagnostic system",
                                                                                        "connecting rod", "battery", "pyrometer"}}
                                                         };

    private GameController gameController;
    private PilotController pilotController;
    private bool problemTimerSet;
    private bool inProblem;
    private float nextProblemTime;
    private float problemTimerStart;

    void Start()
    {
        findGameController();
        findPilotController();
        problemText.text = "";
    }

    void Update()
    {
        if (gameController.getGameOver())
        {
            StopAllCoroutines();
        }
        else
        {
            if (!inProblem)
            {
                if (!problemTimerSet)
                {
                    nextProblemTime = Random.Range(rangeStart, rangeEnd);
                    Debug.Log("Problem timer set: " + nextProblemTime + " seconds");
                    problemTimerStart = Time.time;
                    problemTimerSet = true;
                }
                else
                {
                    if (Time.time - problemTimerStart >= nextProblemTime)
                    {
                        inProblem = true;
                        Debug.Log("Initiating problem state\nStoppping all regular coroutines");
                        gameController.pauseRegularRoutines();
                        StartCoroutine(GenerateProblem(problemSequenceLength));
                    }
                }
            }
        }
    }

    /*
     * Generates a random problem for the player to solve.
     * The problem can have a variable length of 
     */
    public IEnumerator GenerateProblem(int sequenceLength)
    {
        if (sequenceLength <= 0)
        {
            throw new System.Exception("Problems cannot have zero or a negative sequence length");
        }
        else
        {
            List<KeyCode> usableKeys = keysWithNames.Keys.ToList<KeyCode>();            //get usable keys
            KeyCode[] actions = generateUniqueRandomArray(sequenceLength, usableKeys);  //generate unique key sequence
            string o = randomOrdering();                                                //pick a random ordering
            KeyCode[] ordered = sortByOrdering(actions, o);                             //create ordered array
            float startTime = Time.time;
            Debug.Log("Actions: " + KeyCodeArrayToString(actions) + "\nOrdering:" + o + "\nOrdered: " + KeyCodeArrayToString(ordered));

            //In-game dialogue and audio
            StartCoroutine(gameController.switchMusic());
            problemText.text = "What's that weird sound coming from the engine?";
            yield return new WaitForSeconds(4);
            problemText.text = "Uh oh, this isn't looking good!";
            yield return new WaitForSeconds(3);

            //Set problem state for pilot controller
            pilotController.problemState = true;
            //Display problem
            problemText.text = generateProblemText(actions, o);

            //When playing Co-op spawn sharp hazards in the problem state
            if (gameController.coop)
            {
                gameController.enableSharpHazards = true;
            }

            //Problem handler
            int i = 0;
            KeyCode currKey = ordered[i];

            while (i < sequenceLength)
            {
                yield return StartCoroutine(WaitForKey(currKey));

                if (!(++i == sequenceLength))
                {
                    currKey = ordered[i];
                }
            }

            //Determine time bonus and display completion text
            int timeScore = (int) pilotController.dropDownTime - ((int)(Time.time - startTime));
            StartCoroutine(gameController.switchMusic());
            problemText.text = "Looks like the engine is working again! Let's go!" + "\n" + timeScore + " points time bonus!";
            pilotController.problemState = false;

            yield return new WaitForSeconds(3);

            //Add score, increase difficulty and resume regular game
            gameController.AddScore(timeScore);
            gameController.increaseHazardDifficulty();
            increaseProblemDifficulty();
            problemText.text = "";
            gameController.startRegularRoutines();

            inProblem = false;
            problemTimerSet = false;

        }

    }

    /*
     * Coroutine which waits for a key input
     */
    private IEnumerator WaitForKey(KeyCode key)
    {
        while (!Input.GetKeyUp(key))
        {
            yield return null;
        }
        Debug.Log(key + " pressed");
        yield return new WaitForSeconds(0.1f);
    }

    /*
     * Generates an array of unique KeyCodes.
     * Length: length of the random array.
     * KeyList: contains all possible keys.
     * When the length is greater than the length of the list an exception is thrown
     * due the lack of unique values.
     */
    public KeyCode[] generateUniqueRandomArray(int length, List<KeyCode> keyList)
    {
        if (length > keyList.Count)
        {
            throw new System.Exception("Cannot generate unique list by lack of unique values.\n available numbers: " + keyList.Count + ", desired length: " + length + ".");
        }
        else
        {
            List<KeyCode> lst = new List<KeyCode>(keyList);
            KeyCode[] randomNumbers = new KeyCode[length];

            for (int i = 0; i < length; i++)
            {
                int key = Random.Range(0, lst.Count);
                randomNumbers[i] = lst[key];
                lst.RemoveAt(key);
            }
            return randomNumbers;
        }




    }

    /*
     * Sorts an KeyCode array based on given ordering.
     * Order can be: ascending, descending and in order
     */
    public KeyCode[] sortByOrdering(KeyCode[] arr, string order)
    {
        if (order == null)
        {
            throw new System.Exception("order is null");
        }
        else
        {

            switch (order)
            {
                case "ascending":
                    KeyCode[] sortedASC = ShallowCopyKeyCodeArray(arr);
                    System.Array.Sort(sortedASC);
                    return sortedASC;

                case "descending":
                    KeyCode[] sortedDESC = ShallowCopyKeyCodeArray(arr);
                    System.Array.Sort(sortedDESC);
                    System.Array.Reverse(sortedDESC);
                    return sortedDESC;

                case "inorder":
                    return arr;

                default:
                    throw new System.Exception("'" + order + "'" + " is not a valid order");
            }
        }

    }

    /*
     * Generate the problem text using the names in the keysWithNames dictionary
     */
    public string generateProblemText(KeyCode[] arr, string order)
    {
        string s = "Oh no!\n The main engine part '" + ordersWithVerbs[order][Random.Range(0, ordersWithVerbs[order].Length)] + "' is broken!\nQuick! Fix these smaller parts in the right order:\n";
        for (int i = 0; i < arr.Length; i++)
        {
            string keyText = keysWithNames[arr[i]];
            if (i != 0 && i % 3 == 0)
            {
                s += "\n";
            }
            s += keyText + ", ";
        }
        s = s.Remove(s.Length - 2, 2);
        return s;
    }

    /*
     * Increase the problem difficulty based on current sequence length and adjust dropdown time accordingly
     */
    private void increaseProblemDifficulty()
    {
        problemSequenceLength += ((3 / (int)problemSequenceLength));
        pilotController.dropDownTime += 0.75f * (pilotController.dropDownTime / problemSequenceLength);
    }

    //--------------------AUXILIARY FUNCTIONS--------------------\\

    /*
     * Auxiliary function to pick a random ordering
     */
    public string randomOrdering()
    {
        return ordersWithVerbs.Keys.ToArray<string>()[Random.Range(0, 3)];
    }

    /*
     * Auxiliary function that copies all values from a given array into another.
     * The objects themselves remain the same.
     */
    public static KeyCode[] ShallowCopyKeyCodeArray(KeyCode[] arr)
    {
        KeyCode[] shallowCopy = new KeyCode[arr.Length];
        for (int i = 0; i < arr.Length; i++)
        {
            shallowCopy[i] = arr[i];
        }
        return shallowCopy;
    }

    /*
     * Auxiliary function to convert KeyCode arrays to pretty strings
     */
    public static string KeyCodeArrayToString(KeyCode[] arr)
    {
        string s = "{";
        for (int i = 0; i < arr.Length; i++)
        {
            s += arr[i].ToString() + ", ";   
        }
        s = s.Remove(s.Length - 2, 2);
        s += "}";
        return s;
    }

    /*
     * Auxiliary function to find the GameObject with the tag 'GameController' and
     * the attached GameController script.
     */
    public void findGameController()
    {
        GameObject gameControllerObject = GameObject.FindWithTag("GameController");
        if (gameControllerObject != null)
        {
            gameController = gameControllerObject.GetComponent<GameController>();
        }
        else if (gameControllerObject == null)
        {
            Debug.Log("Cannot find 'GameController' script");
        }
    }

    /*
     * Auxiliary function to find the GameObject with the tag 'Player' and
     * the attached PilotController script.
     */
    public void findPilotController()
    {
        GameObject playerObject = GameObject.FindWithTag("Player");
        if (playerObject != null)
        {
            pilotController = playerObject.GetComponent<PilotController>();
        }
        else if (playerObject == null)
        {
            Debug.Log("Cannot find 'PilotController' script");
        }
    }
}
